#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2012 Hiroaki Komatsu
#
import uuid
import datetime
import os
import urllib
from django.utils import simplejson
from google.appengine.api import channel, memcache
from google.appengine.ext import webapp, db
from google.appengine.ext.webapp import util, template

USER_KEY = 'users'
TIME_OUT = 600
name = None

# 開発用サーバかどうかを判定する
def is_dev():
    return os.environ.get('SERVER_SOFTWARE', '').startswith('Development')

# タイムアウトしたユーザーを取り除く
def timeout_check(users):
    now = datetime.datetime.now()
    users = dict([(id, last) for id, last in users.items() if (now - last.datetime) < datetime.timedelta(seconds=TIME_OUT)])
    return users

# 同時接続しているユーザー一覧を取得
def get_users_list(users):
    list = []
    for id in users:
        list.append({
            'client_id' : str(id),
            'ip' : str(users[id].ip),
        })
    return list

# 同時接続しているユーザー一覧を通知する
def push_users(users):
    response = '{"users":' + simplejson.dumps(get_users_list(users), ensure_ascii=False) + '}'
    for id in users:
        # 一人ずつアクセス情報を通知する
        channel.send_message(id, response)

# ユーザークラス
class User:
    def __init__(self, client_id, token):
        self.client_id = client_id
        self.token = token
        self.datetime = datetime.datetime.now()
        self.ip = None
        self.name = None

# 保存データ
class Greeting(db.Model):
    name = db.StringProperty(indexed=False)
    value = db.StringProperty(multiline=True, indexed=False)
    date = db.DateTimeProperty(auto_now_add=True)

# メイン
class MainHandler(webapp.RequestHandler):
    def get(self):
        
        if is_dev():
            script = '/_ah/channel/jsapi'
        else:
            # 本番で「/_ah/channel/jsapi」を読み込むと警告が出るので、本番用はこっちを使う
            script = 'http://talkgadget.google.com/talkgadget/channel.js'
        values = {
            'ip': str(self.request.remote_addr),
            'script': script,
        }
        
        path = os.path.join(os.path.dirname(__file__), 'index.html')
        self.response.out.write(template.render(path, values))

# ログイン
class LoginHandler(webapp.RequestHandler):
    def post(self):
        global name
        
        # クロスドメイン通信を許可
        self.response.headers['Access-Control-Allow-Origin'] = '*'
        self.response.headers['Access-Control-Allow-Headers'] = '*'
        self.response.headers['Access-Control-Allow-Methods'] = 'POST'
        
        # Client IDの作成。64文字以内の一意なASCII文字列であればなんでもいい
        client_id = str(uuid.uuid4())
        
        # Client IDからTokenを作成する。TokenはClient IDと一対一
        token = channel.create_channel(client_id)
        
        # 同時接続しているユーザーのClient ID一覧を取得
        users = memcache.get(USER_KEY)
        if not users:
            users = {}
        
        # タイムアウトしたユーザーを取り除く
        users = timeout_check(users)
        
        # ユーザークラスを作成する
        user = User(client_id, token)
        user.ip = str(self.request.remote_addr)
        
        # 名前
        name = self.request.POST['name']
        try:
            # UTF-8 へのエンコード追加
            name = name.encode('utf-8')
        except UnicodeDecodeError:
            pass
        user.name = name
        
        # 新しいClient IDを追加する
        users[client_id] = user
        memcache.set(USER_KEY, users)
        
        # 同時接続しているユーザー一覧を通知する
        push_users(users)
        
        # データストアを使用
        datas = []
        greetings = db.GqlQuery("SELECT * FROM Greeting ORDER BY date DESC LIMIT 10")
        for i,greeting in enumerate(greetings):
            g_name = "Anon"
            if greeting.name:
                g_name = greeting.name
            
            g_value = ""
            if greeting.value:
                g_value = greeting.value
            
            datas.append({
                'name' : urllib.quote(g_name.encode('utf-8')),
                'value' : urllib.quote(g_value.encode('utf-8')),
                'datetime' : str(greeting.date),
            })
        datas.reverse()
        
        # レスポンス
        response = {
            'client_id': client_id,
            'token': token,
            'users': get_users_list(users),
            'datas': datas,
        }
        
        self.response.headers['Content-Type'] = 'text/plain;charset=UTF-8'
        self.response.out.write(response)

# メッセージ受信
class PostHandler(webapp.RequestHandler):
    def post(self):
        global name
        
        # クロスドメイン通信を許可
        self.response.headers['Access-Control-Allow-Origin'] = '*'
        self.response.headers['Access-Control-Allow-Headers'] = '*'
        self.response.headers['Access-Control-Allow-Methods'] = 'GET, POST, OPTIONS'
        
        client_id = str(self.request.POST['client_id'])
        value = self.request.POST['value']
        ip = str(self.request.remote_addr)
        try:
            # UTF-8 へのエンコード追加
            value = value.encode('utf-8');
        except UnicodeDecodeError:
            pass
        
        now = str(datetime.datetime.now(JST()))
        message = {
            'value': value,
            'client_id': str(client_id),
            'datetime': now,
            'ip': str(ip),
            'name': name,
        }
        response = '{"message":' + simplejson.dumps(message, ensure_ascii=False) + '}'
        
        # 同時接続中ユーザーのClient ID一覧を取得
        users = memcache.get(USER_KEY)
        if client_id in users:
            
            # タイムアウトしたユーザーを取り除く
            users = timeout_check(users)
            users[client_id].datetime = datetime.datetime.now()
            memcache.set(USER_KEY, users)
            
            for id in users:
                # 一人ずつ更新を通知する
                channel.send_message(id, response)
            
            # データストアに保存
            greeting = Greeting()
            if name:
                greeting.name = db.Text(name, encoding="utf-8")
            greeting.value = db.Text(value, encoding="utf-8")
            greeting.ip = ip
            greeting.put()
        
        # レスポンス
        self.response.out.write(response)

# Channel APIの接続
class ChannelConnectHandler(webapp.RequestHandler):
    def post(self):
        pass

# Channel APIの切断
class ChannelDisconnectHandler(webapp.RequestHandler):
    def post(self):
        client_id = self.request.get('from')
        
        # 切断されたユーザーを取り除く
        users = memcache.get(USER_KEY)
        users = dict([(id, last) for id, last in users.items() if last.client_id != client_id])
        memcache.set(USER_KEY, users)
        
        # 同時接続しているユーザー一覧を通知する
        push_users(users)

# 日本時間
class JST(datetime.tzinfo):
    def utcoffset(self, dt):
        return datetime.timedelta(hours=9)
    def dst(self, dt):
        return datetime.timedelta(0)
    def tzname(self, dt):
        return "JST"

def main():
    application = webapp.WSGIApplication([('/', MainHandler),('/login', LoginHandler),('/post', PostHandler),('/_ah/channel/connected/', ChannelConnectHandler),('/_ah/channel/disconnected/', ChannelDisconnectHandler)],debug=True)
    util.run_wsgi_app(application)

if __name__ == '__main__':
    main()

